# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.web import server, static, resource
from twisted.internet import reactor
from twisted.application import service
from twisted.web.util import Redirect
from twisted.web.woven.guard import UsernamePasswordWrapper, SessionWrapper
from xml.sax.saxutils import quoteattr
from yarn import remote, errors, auth, url, features
import os, urllib
import hep

class WebServerService(service.Service):
    PORT = 5080
    NAME = 'http'
    
    def startService(self):
        # register an IResource interface with the data manager
        self.dm = self.parent.getServiceNamed('Data Manager')
        self.dm.avatarInterfaces[resource.IResource] = self.getResource
        
        self.factory = getFactory(self.parent)
        reactor.listenTCP(self.PORT, self.factory)
        
    def getResource(self, user):
        if user:
            return UserResource(user)
        else:
            anonUser = self.dm.users['anonymous']
            return UserResource(anonUser)

def getFactory(app):
    portal = app.getServiceNamed('Authenticator').portal    
    res =  SessionWrapper(UsernamePasswordWrapper(
        portal, lambda _: Redirect('/')))
    site = server.Site(res)
    return site

class NotFoundPage(resource.Resource):
    isLeaf = 1

    def __init__(self, user):
        resource.Resource.__init__(self)    

    def render(self, request):
        request.status = 404
        return "File Not Found."

class UserResource(resource.Resource):

    def __init__(self, user):
        self.user = user
        resource.Resource.__init__(self)
        # add static path
        staticFolder = os.path.join(os.path.split(hep.__file__)[0],
                                     'web', 'static')
        self.putChild('static', static.File(staticFolder))
        self.putChild('', MainPage(self.user))
        self.putChild('setup', SetupPage(self.user))
        
class UserPage(resource.Resource):
    
    def __init__(self, user):
        self.user = user
        resource.Resource.__init__(self)

    def redirectIfAnonymous(self, request):
        if self.user.name == 'anonymous':
            request.redirect('/')
            request.finish()
            return True
        else:
            return False


class MainPage(UserPage):
    def render(self, request):
        if self.user.name != 'anonymous':
            request.redirect('/setup');
            return ''

        return """<!DOCTYPE html PUBLIC "-//W3C/DTD XHTML 1.0 Strict//EN"
	  "http://www.w3c.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head>
    <title>Hep: Login</title>
    <link rel='stylesheet' type='text/css' href='/static/hep.css' />
  </head>
  <body>
    <h1><span id='hep'>hep</span> message server</h1>
    <div class='instructions'>Please log in.</div>
    <form id='login' action='perspective-init' method='post'>
      <table>
	<tr>
	  <td class='label'>User Name:</td>
	  <td><input type='text' name='username' /></td>
	</tr>
	<tr>
	  <td class='label'>Password:</td>
	  <td><input type='password' name='password' /></td>
	</tr>
      </table>
      <div class='submitbuttons'><input type='submit' value='OK' /></div>
    </form>
  </body>
</html> """
        
class SetupPage(UserPage):
    
    def __init__(self, user):
        UserPage.__init__(self, user)
        self.putChild('add', AddSitePage(self.user))
        self.putChild('edit', EditSitePage(self.user))
        self.putChild('delete', DeleteSitePage(self.user))
        self.putChild('folders', DefaultWritePage(self.user))
        self.putChild('sites', SitesPage(self.user))
    
    def render(self, request):
        self.redirectIfAnonymous(request) or request.redirect(
            '/setup/sites')
        return ''

class SitesPage(UserPage):

    def render(self, request):
        if self.redirectIfAnonymous(request): return ''
        
        request.setHeader('Content-Type', 'text/html; charset=utf-8')
        request.write("""
        <!DOCTYPE html PUBLIC "-//W3C/DTD XHTML 1.0 Strict//EN"
	  "http://www.w3c.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
          <html>
          <head>
            <title>Hep: Setup: Sites</title>
            <link rel='stylesheet' type='text/css' href='/static/hep.css' />
          </head>
          <body>
          <h1><span id='hep'>hep</span> message server</h1>
            <ul id='nav'>
              <li><a href='/setup/sites' class='currentpage'>Sites</a></li>
              <li><a href='/setup/folders'>Special Folders</a></li>
            </ul>
            <h2>Sites</h2>
            <form action='/setup/add' method='get'>
              Add new site:
              <input type='text' name='url' size='50'>
              <input type='submit' value='Go'>
            </form>
        """)

        sitesFolder = self.user.view.subFolders['Sites']
        if sitesFolder.links:
            request.write("""
            <table>
              <thead>
                <tr><th>Name</th><th colspan='2'>URL</th></tr>
              </thead>
              <tbody>
            """)
            for path, link in sitesFolder.links.items():
                request.write("""
                <tr>
                <td>%s</td>
                <td>%s</td>
                <td><a href='/setup/edit?site=%s' class='edit'>Edit</a>
                    <a href='/setup/delete?site=%s' class='delete'>Delete</a>
                </td>
                </tr>
                """ % (path.encode('utf-8'), link.url.encode('utf-8'),
                       urllib.quote(path.encode('utf-8')),
                       urllib.quote(path.encode('utf-8'))))
            request.write("""
            </tbody>
            </table>
            """)

        request.write("""
        </body>
        </html>""")
        return ""


class DefaultWritePage(UserPage):

    def render(self, request):
        if self.redirectIfAnonymous(request): return ''

        # set folder if it was passed
        folder = request.args.get('folder', [''])[0]
        if folder:
            self.user.config['weblogfolder'] = folder
        request.setHeader('Content-Type', 'text/html; charset=utf-8')
        request.write("""
        <!DOCTYPE html PUBLIC "-//W3C/DTD XHTML 1.0 Strict//EN"
	  "http://www.w3c.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
          <html>
          <head>
            <title>Hep: Setup: Special Folders</title>
            <link rel='stylesheet' type='text/css' href='/static/hep.css' />
          </head>
          <body>
          <h1><span id='hep'>hep</span> message server</h1>
            <ul id='nav'>
              <li><a href='/setup/sites'>Sites</a></li>
              <li><a href='/setup/folders' class='currentpage'>
                Special Folders</a></li>
            </ul>
          <h2>Reply-To Site</h2>
          <div class='instructions'>
          This should be your primary weblog, or the site you generally send
          public messages to.
          </div>
          <div class='instructions'>
          Hep will use this site as the default 'Reply-To' address for
          incoming messages.
          </div>
        """)
        writableFolders = self.getWritableFolders()
        if writableFolders:
            request.write("""
            <form>
            <table>
            <thead>
            <tr><th>Name</th><th>URL</th></tr>
            </thead>
            <tbody>
            """)
            currentFolder = self.user.config.get('weblogfolder', '')
            for folder in writableFolders:
                fName = folder.name.encode('utf-8')
                request.write("<tr>")
                request.write("""
                <td>
                <input type='radio' name='folder'
                 value=%s id=%s %s />
                <label for=%s>%s</label>
                </td>
                """ % (quoteattr(folder.url), quoteattr('folder_' + fName),
                       (currentFolder == folder.url) and 'checked' or '',
                       quoteattr('folder_' + fName), fName))
                request.write("<td>%s</td>" % (folder.url.encode('utf-8')))
                request.write("</tr>")
            request.write("""
            </tbody>
            </table>
            <div class='submitbuttons'>
            <input type='submit' value='Save' />
            </div>
            </form>
            """)
        else:
            request.write("""
            <div class='instructions'>
            (No writable folders were found in your current sites.)
            </div>""")
        request.write("</body></html>")
        return ''
    
    def getWritableFolders(self, parent=None):
        folders = []
        if not parent: parent = self.user.view.subFolders['Sites']
        for f in parent.subFolders.values():
            if not f.connected: continue # skip disconnected folders
            if features.POST_MESSAGE in f.features:
                folders.append(f)
            folders = folders + self.getWritableFolders(f)
        return folders

class SiteActionPage(UserPage):
    header =  """
        <!DOCTYPE html PUBLIC "-//W3C/DTD XHTML 1.0 Strict//EN"
	  "http://www.w3c.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
          <html>
          <head>
            <title>Hep: Sites: Add Site</title>
            <link rel='stylesheet' type='text/css' href='/static/hep.css' />
          </head>
          <body>
            <h1><span id='hep'>hep</span> message server</h1>
            <ul id='nav'>
              <li><a href='/setup/sites' class='currentpage'>Sites</a></li>
              <li><a href='/setup/folders'>
                Special Folders</a></li>
            </ul>
            """


class AddSitePage(SiteActionPage):

    def render(self, request):
        if self.redirectIfAnonymous(request): return ''
        
        u = request.args.get('url', [''])[0]
        username = request.args.get('site_user', [''])[0]
        password = request.args.get('site_pass', [''])[0]
        if username or password:
            creds = auth.UserPassCredentials(username, password)
        else:
            creds = None
        self.testConnection(u, request, creds)
        return server.NOT_DONE_YET

    def testConnection(self, url_, request, creds=None):
        if creds:
            keys = auth.Keychain()
            keys[url_] = creds
        else:
            keys = self.user.keychain
        v = remote.Folder(keychain=keys)
        v.connect(url_, self.user.connector).addCallback(
            self._connectComplete, v, request).addErrback(
            self._connectComplete, v, request, False)

    def _connectComplete(self, result, folder, request, succeeded=True):
        pageParts = [self.header]
        pageParts.append("""
        <form>
        <input type='hidden' name='url' value='%s'>
        """ % (
            folder.url
            ))
        if succeeded:
            # save keychain info
            creds = folder.keychain.get(folder.url)
            if creds:
                self.user.keychain[folder.url] = creds
                self.user.keychain.save()

            # check if this folder contains messages, or just related urls.
            if features.MESSAGES in folder.features or (
                features.FOLDERS in folder.features):
                link = url.Link(folder.name, folder.url, "alternate")
                self.addLink(link, request)
            else:
                folder.getAlternateLinks().addCallback(
                    self.showFoundLinks, request).addErrback(
                    self._getLinksFailed, request)
            return
        elif result.check(errors.CredentialsRequired, errors.LoginFailed):
            pageParts.append("""
            This site requires a user name and password. If
            you know a valid user name and password, enter them below;
            otherwise, press the Back button to go to the previous page.
            <table>
            <tr><td>Username: </td>
                <td><input type='text' name='site_user' /></td></tr>
            <tr><td>Password: </td>
                <td><input type='password' name='site_pass' /></td></tr>
            </table>
            <input type='submit' />
            """)
        else:
            pageParts.append("""
            <div class='instructions'>
            Hep was unable to establish a connection to the site you entered:
            <br />%s
            <p>
            %s
            </p>
            </div>
            """ % (folder.url, result.getErrorMessage()))
                          
        pageParts.append("</form>")
        request.write("".join(pageParts))
        request.finish()

    def showFoundLinks(self, links, request):
        if not links:
            request.write(self.header)
            request.write("""
            <div class='instructions'>
            Hep couldn't find any message sources at the URL you entered:<br />
            %s
            <p>
            If you were expecting to find an RSS feed on that site, it may
            be that the site's main page didn't contain a link to the feed.
            Try entering the URL of the feed itself.
            </p>
            </div>
            </body>
            </html>
            """ % request.args['url'][0])
            request.finish()
            return
        
        elif len(links) == 1:
            self.testConnection(links[0].url, request)
            return

        request.write(self.header)
        request.write("""
        <div class='instructions'>
        Hep found multiple sites at the URL '%s'. <br />
        Please select one of the following options:
        </div>
        <form>
        """ % request.args['url'][0])
        for link in links:
            u = link.url.encode('utf-8')
            t = link.title.encode('utf-8')
            request.write("""
            <div>
            <input type='radio' name='url' value='%s' id='url_%s' %s />
            <label for='url_%s'>%s <span class='wee'>(%s)</span></label>
            </div>
            """ % (u, u, links.index(link) or "checked='checked'",
                   u, t, u))
        request.write("""
        <div class='submitbuttons'>
        <input type='submit' value='Add Site' />
        </div>
        </form>
        </body>
        </html>
        """)
        request.finish()

    def _getLinksFailed(self, err, request):
        request.write(self.header + """
        <div class='instructions'>
        Loading links failed: %s
        </div>
        </body>
        </html>""" % err)
        request.finish()    

    def addLink(self, link, request):
        self.user.view.subFolders['Sites'].createLink(link).addCallback(
            self._addSuccessful, request).addErrback(self._addFailed, request)

    def _addSuccessful(self, path, request):
        sitesFolder = self.user.view.subFolders['Sites']
        sitesFolder.subFolders[path].refresh().addCallback(
            self._refreshCompleted, request).addErrback(
            self._refreshCompleted)

    def _refreshCompleted(self, result, request):
        request.redirect('/setup')
        request.finish()

    def _addFailed(self, error, request):
        request.write(self.header + """
        <div class='information'>
        Unable to add to Sites: %s
        </div>
        </body>
        </html>
        """ % error)
        request.finish()
                                
class DeleteSitePage(SiteActionPage):

    def render(self, request):
        if self.redirectIfAnonymous(request): return ''
        siteName = request.args['site'][0]
        sites = self.user.view.subFolders['Sites']
        
        if request.args.get('verified'):
            sites.deleteFolder(siteName).addCallback(
                self._deletedSite, request).addErrback(
                self._deleteFailed, request)
            return server.NOT_DONE_YET
        elif request.args.get('cancel'):
            request.redirect('/setup/sites')
            return ''
        else:
            return self.header + """
            <h2>Delete Site</h2>
            <div class='instructions'>
            Do you really want to remove "%s" from your sites?
            </div>
            <form>
            <input type='hidden' name='site' value=%s />
            <input type='submit' name='verified' value='Yes' />
            <input type='submit' name='cancel' value='No' />
            </form>
            </body></html>
            """ % (siteName, quoteattr(siteName))

    def _deletedSite(self, result, request):
        request.redirect('/setup/sites')
        request.finish()

    def _deleteFailed(self, err, request):
        request.write(self.header + """
        <div class='error'>
        Delete Failed: %s
        </div>
        </body>
        </html>
        """ % err)
        request.finish()
    

class EditSitePage(SiteActionPage):
    
    def render(self, request):
        if self.redirectIfAnonymous(request):
            return ''
        elif request.args.get('cancel'):
            request.redirect('/setup/sites')
            return ''
        elif request.args.get('save'):
            self.saveSite(request)
            return server.NOT_DONE_YET
        
        siteName = request.args['site'][0]
        link = self.user.view.subFolders['Sites'].links[siteName]
        request.setHeader('Content-Type', 'text/html; charset=utf-8')
        return self.header + """
        <h2>Edit Site</h2>
          <form>
          <input type='hidden' name='site' value=%s />
          <table>
            <tr>
              <td class='label'>Name:</td>
              <td><input type='text' name='name' value=%s /></td>
            </tr>
            <tr>
              <td class='label'>URL:</td>
              <td><input type='text' name='url' size='50' value=%s /></td>
            <tr>
          </table>
          <div class='submitbuttons'>
            <input type='submit' value='OK' name='save' />
            <input type='submit' value='Cancel' name='cancel' />
          </div>
          </form>
        </body>
        </html>
        """ % (quoteattr(siteName.encode('utf-8')),
               quoteattr(siteName.encode('utf-8')),
               quoteattr(link.url))

    def saveSite(self, request):
        siteName = request.args['site'][0]
        sites = self.user.view.subFolders['Sites']
        link = sites.links[siteName]
        name = request.args.get('name', [''])[0]
        u = request.args.get('url', [''])[0]
        if not name or not u:
            raise Exception, "Missing name or URL" # fixme: handle this nicer
        newLink = url.Link(name, u, link.rel, link.mimetype)
        sites.editLink(link, newLink).addCallback(
            self._savedSite, request).addErrback(
            self._saveFailed, request)

    def _savedSite(self, result, request):
        request.redirect('/setup/sites')
        request.finish()

    def _saveFailed(self, error, request):
        request.write("error: %s" % error.getErrorMessage())
        request.finish()
