# This file is part of Hep, a multiprotocol message server.
# Copyright (C) 2004 Abe Fettig
#
# Hep is free software; you can redistribute it and/or
# modify it under the terms of version 2 of the GNU General
# Public License as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.mail import pop3
from twisted.internet import protocol, reactor, defer
from twisted.application import service
from yarn import message, url

class Service(service.Service):
    PORT = 110
    NAME = 'pop3'
    
    def startService(self):
        # add support for getting POP accounts out of the datamanager service
        dm = self.parent.getServiceNamed('Data Manager')
        dm.avatarInterfaces[pop3.IMailbox] = openPOPMailbox
        
        portal = self.parent.getServiceNamed('Authenticator').portal
        self.factory = POP3Factory(portal)
        self.factory.app = self.parent
        
        reactor.listenTCP(self.PORT, self.factory)

class POP3Factory(protocol.Factory):

    def __init__(self, portal):
        self.portal = portal

    def buildProtocol(self, address):
        protocol = pop3.POP3()
        protocol.portal = self.portal
        return protocol

def openPOPMailbox(user):
    box = HepPOPMailbox(user)
    return box.load().addCallback(lambda _: box)

class HepPOPMailbox(object):
    # FIXME: holding every single unread message in memory probably isn't
    # the best implementation....
    
    __implements__ = pop3.IMailbox

    uidlSep = "|"

    def __init__(self, user):
        self.user = user
        self.messages = {}
        self.deleted = []

    def keys(self):
        keys = self.messages.keys()
        keys.sort()
        return keys

    def getMessageByKey(self, key):
        return self.messages[key][-1]

    def getMessageByIndex(self, index):
        return self.getMessageByKey(self.keys()[index])

    def load(self):
        # crawl view and build map of messages.
        return self.crawl(self.user.view)

    def crawl(self, view):
        actions = []
        for folder in view.subFolders.values():
            actions.append(self.crawl(folder))
        d = defer.Deferred()
        view.listMessages().addCallback(self._gotMessageList, view, d)
        actions.append(d)
        return defer.DeferredList(actions)

    def _gotMessageList(self, messageKeys, folder, finished):
        getters = []
        "POP: gotMessageList", messageKeys
        for key in messageKeys:
            if not folder.messageMetadata[key].get('seen'):
                print "POP: downloading message"
                getters.append(folder.getMessage(key).addCallback(
                    self._gotMessage, key, folder))
        defer.DeferredList(getters).chainDeferred(finished)

    def _gotMessage(self, msg, key, folder):
        uidl = self.uidlSep.join([str(folder.url), key])
        msgCopy = self.user.getMessageForEmail(msg)
        self.messages[uidl] = (folder, key,
                               str(msgCopy))

    def listMessages(self, index=None):
        if index:
            return len(self.getMessageByIndex(index))
        else:
            return [len(self.getMessageByKey(key)) for  key in self.keys()]

    def getMessage(self, index):
        from cStringIO import StringIO
        return StringIO(self.getMessageByIndex(index))

    def getUidl(self, index):
        return self.keys()[index]

    def deleteMessage(self, index):
        self.deleted.append(index)

    def undeleteMessage(self, index):
        self.deleted.remove(index)

    def sync(self):
        for mIndex in self.deleted:
            folder, key, msg = self.messages[self.keys()[mIndex]]
            folder.messageMetadata[key]['seen'] = 1
        self.user.view.saveSoon()
    
